@@ -6,7 +6,7 @@
 #ifndef _IP_VS_H
 #define _IP_VS_H
 
-#define IP_VS_VERSION_CODE            0x000101
+#define IP_VS_VERSION_CODE            0x000102
 #define NVERSION(version)                       \
         (version >> 16) & 0xFF,                 \
         (version >> 8) & 0xFF,                  \
@@ -104,6 +104,8 @@ struct ip_vs_rule_user {
 #include <linux/skbuff.h>               /* for struct sk_buff */
 #include <linux/ip.h>                   /* for struct iphdr */
 #include <asm/atomic.h>                 /* for struct atomic_t */
+#include <linux/netdevice.h>		/* for struct neighbour; */
+#include <net/dst.h>			/* for struct dst_entry */
 
 
 /*  #define CONFIG_IP_VS_DEBUG */
@@ -174,7 +176,8 @@ enum {
 	NET_IPV4_VS_TO_UDP=17,
 	NET_IPV4_VS_TO_ICMP=18,
 	NET_IPV4_VS_LBLC_EXPIRE=19,
-	NET_IPV4_VS_LBLCR_EXPIRE=19,
+	NET_IPV4_VS_LBLCR_EXPIRE=20,
+	NET_IPV4_VS_USE_DST=21,
 };
 
 enum {
@@ -280,7 +283,7 @@ struct ip_vs_service {
 
 /*
  *	The real server destination forwarding entry
- *	with ip address, port
+ *	with ip address, port number, and so on.
  */
 struct ip_vs_dest {
 	struct list_head	n_list;   /* d-linked list head */
@@ -292,7 +295,10 @@ struct ip_vs_dest {
 	atomic_t		activeconns;	/* active connections */
 	atomic_t		inactconns;	/* inactive connections */
 	atomic_t		refcnt; 	/* reference counter */
-	struct list_head	d_list;   /* table with all dests */
+	struct list_head	d_list;         /* table with all dests */
+	spinlock_t		dst_lock;	/* lock dst_cache */
+	struct dst_entry	*dst_cache;	/* destination cache entry */
+	u32			dst_rtos;	/* RT_TOS(tos) for dst */
 
 	/* for virtual service */
 	__u16			protocol; /* which protocol (TCP/UDP) */
@@ -333,7 +339,6 @@ struct ip_vs_app
         char                    *name;    /* name of application module */
         unsigned                type;     /* type = proto<<16 | port
                                              (host byte order)*/
-        /*  atomic_t                refcnt; */   /* reference counter */
 	struct module		*module;  /* THIS_MODULE/NULL */
         
         /* ip_vs_app initializer */
@@ -558,6 +563,44 @@ extern __inline__ char ip_vs_fwd_tag(struct ip_vs_conn *cp)
 }
 
 
+/*
+ *      Destination cache
+ */
+static inline void
+__ip_vs_dst_set(struct ip_vs_dest *dest, u32 rtos, struct dst_entry *dst)
+{
+	struct dst_entry *old_dst;
+
+	old_dst = dest->dst_cache;
+	dest->dst_cache = dst;
+	dest->dst_rtos = rtos;
+	dst_release(old_dst);
+}
+
+static inline void
+__ip_vs_dst_reset(struct ip_vs_dest *dest)
+{
+	struct dst_entry *old_dst;
+
+	old_dst = dest->dst_cache;
+	dest->dst_cache = NULL;
+	dst_release(old_dst);
+}
+
+static inline struct dst_entry *
+__ip_vs_dst_check(struct ip_vs_dest *dest, u32 rtos, u32 cookie)
+{
+	struct dst_entry *dst = dest->dst_cache;
+
+	if (dst && (dst->obsolete || rtos != dest->dst_rtos) &&
+		dst->ops->check(dst, cookie) == NULL) {
+		dest->dst_cache = 0;
+		return NULL;
+	}
+	return dst;
+}
+
+
 #endif /* __KERNEL__ */
 
 #endif	/* _IP_VS_H */
