@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.8 2000/11/26 04:35:39 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.9 2000/12/19 15:04:23 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -49,6 +49,7 @@
 #include "ip_vs.h"
 //#include <net/ip_vs.h>
 
+extern int sysctl_use_dst;
 
 EXPORT_SYMBOL(register_ip_vs_scheduler);
 EXPORT_SYMBOL(unregister_ip_vs_scheduler);
@@ -468,8 +469,10 @@ static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 {
         struct rtable *rt;			/* Route to the other host */
+	struct ip_vs_dest *dest;
 	struct iphdr  *iph = skb->nh.iph;
 	u8     tos = iph->tos;
+	u32    rtos;
 	union ip_vs_tphdr h;
 	unsigned short size;
 	int doff;
@@ -524,11 +527,22 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
                 IP_VS_DBG(1, "filled cport=%d\n", ntohs(cp->dport));
         }
                 
-        if (ip_route_output(&rt, cp->daddr, 0, RT_TOS(tos), 0)) {
-		IP_VS_DBG(0, "ip_vs_nat_xmit(): route error, dest: %08X\n",
-                          cp->daddr);
-		goto tx_error_icmp;
+	dest = cp->dest;
+	rtos = RT_TOS(tos);
+	spin_lock(&dest->dst_lock);
+	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
+		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
+			spin_unlock(&dest->dst_lock);
+			IP_VS_DBG(0, "ip_vs_nat_xmit(): route error, "
+				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
+			goto tx_error_icmp;
+		}
+		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
+		IP_VS_DBG(1, "NAT: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
+			NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
+			rtos);
 	}
+	spin_unlock(&dest->dst_lock);
 
         /* MTU checking */
         mtu = rt->u.dst.pmtu;
@@ -640,14 +654,14 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 static int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 {
 	struct rtable *rt;			/* Route to the other host */
+	struct ip_vs_dest *dest;
 	struct net_device *tdev;		/* Device to other host */
 	struct iphdr  *old_iph = skb->nh.iph;
 	u8     tos = old_iph->tos;
+	u32    rtos;
 	u16    df = old_iph->frag_off;
 	struct iphdr  *iph;			/* Our new IP header */
 	int    max_headroom;			/* The extra header space needed */
-	u32    dst = cp->daddr;
-	u32    src = 0;
 	int    mtu;
 
 	EnterFunction(4);
@@ -659,11 +673,23 @@ static int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 		goto tx_error;
 	}
 
-	if (ip_route_output(&rt, dst, src, RT_TOS(tos), 0)) {
-		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): route error, "
-                          "dest: %u.%u.%u.%u", NIPQUAD(dst));
-		goto tx_error_icmp;
+	dest = cp->dest;
+	rtos = RT_TOS(tos);
+	spin_lock(&dest->dst_lock);
+	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
+		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
+			spin_unlock(&dest->dst_lock);
+			IP_VS_DBG(0, "ip_vs_tunnel_xmit(): route error, "
+				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
+			goto tx_error_icmp;
+		}
+		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
+		IP_VS_DBG(1, "TUN: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
+			NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
+			rtos);
 	}
+	spin_unlock(&dest->dst_lock);
+
 	tdev = rt->u.dst.dev;
 
 	mtu = rt->u.dst.pmtu - sizeof(struct iphdr);
@@ -760,17 +786,30 @@ static inline int do_ip_send(struct sk_buff *skb)
 static int ip_vs_dr_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 {
 	struct rtable *rt;			/* Route to the other host */
+	struct ip_vs_dest *dest;
 	struct iphdr  *iph = skb->nh.iph;
 	u8     tos = iph->tos;
+	u32    rtos;
         int    mtu;
 
 	EnterFunction(4);
 
-        if (ip_route_output(&rt, cp->daddr, 0, RT_TOS(tos), 0)) {
-		IP_VS_DBG(0, "ip_vs_dr_xmit(): route error, "
-                          "dest: %u.%u.%u.%u\n", NIPQUAD(cp->daddr));
-		goto tx_error_icmp;
+	dest = cp->dest;
+	rtos = RT_TOS(tos);
+	spin_lock(&dest->dst_lock);
+	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
+		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
+			spin_unlock(&dest->dst_lock);
+			IP_VS_DBG(0, "ip_vs_dr_xmit(): route error, "
+				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
+			goto tx_error_icmp;
+		}
+		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
+		IP_VS_DBG(1, "DR: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
+			NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
+			rtos);
 	}
+	spin_unlock(&dest->dst_lock);
 
         /* MTU checking */
         mtu = rt->u.dst.pmtu;
@@ -1152,7 +1191,9 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	unsigned short   len   = ntohs(iph->tot_len) - (iph->ihl * 4);
 	struct ip_vs_conn *cp;
         struct rtable *rt;			/* Route to the other host */
+	struct ip_vs_dest *dest;
         u8     tos = iph->tos;
+	u32    rtos;
         int    mtu;
 
  	IP_VS_DBG(2, "icmp in (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
@@ -1242,11 +1283,22 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
         /*
          * Send the mangled packet immediately here
          */
-        if (ip_route_output(&rt, cp->daddr, 0, RT_TOS(tos), 0)) {
-                IP_VS_DBG(0, "ip_vs_in_imcp(): route error, dest: %08X\n",
-                          cp->daddr);
-                goto tx_error_icmp;
+	dest = cp->dest;
+	rtos = RT_TOS(tos);
+	spin_lock(&dest->dst_lock);
+	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
+		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
+			spin_unlock(&dest->dst_lock);
+			IP_VS_DBG(0, "ip_vs_in_imcp(): route error, "
+				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
+			goto tx_error_icmp;
+		}
+		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
+		IP_VS_DBG(1, "ICMP: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
+			NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
+			rtos);
         }
+	spin_unlock(&dest->dst_lock);
 
         /* MTU checking, is it necessary here??? */
         mtu = rt->u.dst.pmtu;
