@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_ctl.c,v 1.6 2000/11/26 04:35:39 wensong Exp $
+ * Version:     $Id: ip_vs_ctl.c,v 1.7 2000/12/19 15:04:23 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -72,6 +72,7 @@ static int sysctl_ip_vs_drop_packet = 0;
 static int sysctl_ip_vs_secure_tcp = 0;
 static int sysctl_ip_vs_amemthresh = 1024;
 static int sysctl_ip_vs_am_droprate = 10;
+static int sysctl_use_dst = 1;
 
 #ifdef CONFIG_IP_VS_DEBUG
 static int sysctl_ip_vs_debug_level = 0;
@@ -570,6 +571,7 @@ ip_vs_get_trash_dest(struct ip_vs_service *svc, __u32 daddr, __u16 dport)
 				  NIPQUAD(dest->addr), ntohs(dest->port));
 			e = e->prev;
 			list_del(&dest->n_list);
+			__ip_vs_dst_reset(dest);
 			kfree(dest);
 		}
 	}
@@ -581,11 +583,16 @@ ip_vs_get_trash_dest(struct ip_vs_service *svc, __u32 daddr, __u16 dport)
 /*
  *  Update a destination in the given service
  */
-static void __ip_vs_update_dest(struct ip_vs_service *svc,
+static int __ip_vs_update_dest(struct ip_vs_service *svc,
 				struct ip_vs_dest *dest,
 				struct ip_vs_rule_user *ur)
 {
 	int conn_flags;
+	unsigned atype;
+
+	atype = inet_addr_type(ur->daddr);
+	if (atype != RTN_LOCAL && atype != RTN_UNICAST)
+		return -EINVAL;
 
 	/*
 	 *    Set the weight and the flags
@@ -598,7 +605,7 @@ static void __ip_vs_update_dest(struct ip_vs_service *svc,
 	/*
 	 *    Check if local node and update the flags
 	 */
-	if (inet_addr_type(ur->daddr) == RTN_LOCAL) {
+	if (atype == RTN_LOCAL) {
 		conn_flags = (conn_flags & ~IP_VS_CONN_F_FWD_MASK)
 			| IP_VS_CONN_F_LOCALNODE;
 	}
@@ -624,24 +631,28 @@ static void __ip_vs_update_dest(struct ip_vs_service *svc,
 	 *    Set the dest status flags
 	 */
 	dest->flags |= IP_VS_DEST_F_AVAILABLE;
+
+	return 0;
 }
 
         
 /*
  *  Create a destination for the given service
  */
-static struct ip_vs_dest *
-ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
+static int
+ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur,
+	struct ip_vs_dest **destp)
 {
 	struct ip_vs_dest *dest;
+	int ret;
 
 	EnterFunction(4);
 
-	dest = (struct ip_vs_dest*) kmalloc(sizeof(struct ip_vs_dest),
+	*destp = dest = (struct ip_vs_dest*) kmalloc(sizeof(struct ip_vs_dest),
 					GFP_ATOMIC);
 	if (dest == NULL) {
 		IP_VS_ERR("ip_vs_new_dest: kmalloc failed.\n");
-		return NULL;
+		return -ENOMEM;
 	}
 	memset(dest, 0, sizeof(struct ip_vs_dest));
 
@@ -657,11 +668,13 @@ ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 	atomic_set(&dest->refcnt, 0);
 
 	INIT_LIST_HEAD(&dest->d_list);
-	__ip_vs_update_dest(svc, dest, ur);
+	dest->dst_lock = SPIN_LOCK_UNLOCKED;
+	ret = __ip_vs_update_dest(svc, dest, ur);
+	if (ret) kfree(dest);
 
 	LeaveFunction(4);
 
-	return dest;
+	return ret;
 }
 
 
@@ -674,6 +687,7 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 	struct ip_vs_dest *dest;
 	__u32 daddr = ur->daddr;
 	__u16 dport = ur->dport;
+	int ret;
 
 	EnterFunction(4);
 
@@ -705,7 +719,10 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 			NIPQUAD(dest->vaddr),
 			ntohs(dest->vport));
 
-		__ip_vs_update_dest(svc, dest, ur);
+		ret = __ip_vs_update_dest(svc, dest, ur);
+		if (ret)
+			return ret;
+
 		/*
 		 * Get the destination from the trash
 		 */
@@ -727,10 +744,9 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 	/*
 	 * Allocate and initialize the dest structure
 	 */
-	dest = ip_vs_new_dest(svc, ur);
-	if (dest == NULL) {
-		IP_VS_ERR("ip_vs_add_dest(): out of memory\n");
-		return -ENOMEM;
+	ret = ip_vs_new_dest(svc, ur, &dest);
+	if (ret) {
+		return ret;
 	}
 
 	/*
@@ -764,6 +780,7 @@ static int ip_vs_edit_dest(struct ip_vs_service *svc,
 	struct ip_vs_dest *dest;
 	__u32 daddr = ur->daddr;
 	__u16 dport = ur->dport;
+	int ret;
 
 	EnterFunction(4);
 
@@ -781,11 +798,11 @@ static int ip_vs_edit_dest(struct ip_vs_service *svc,
 		return -ENOENT;
 	}
 
-	__ip_vs_update_dest(svc, dest, ur);
+	ret = __ip_vs_update_dest(svc, dest, ur);
 
         LeaveFunction(4);
         
-	return 0;
+	return ret;
 }
 
 
@@ -807,9 +824,10 @@ static void __ip_vs_del_dest(struct ip_vs_dest *dest)
 	 *  if nobody refers to it (refcnt=0). Otherwise, throw
 	 *  the destination into the trash.
 	 */
-	if (atomic_dec_and_test(&dest->refcnt))
+	if (atomic_dec_and_test(&dest->refcnt)) {
+		__ip_vs_dst_reset(dest);
 		kfree(dest);
-	else {
+	} else {
 		IP_VS_DBG(1, "Moving dest %u.%u.%u.%u:%u into trash, refcnt=%d\n",
                           NIPQUAD(dest->addr), ntohs(dest->port),
                           atomic_read(&dest->refcnt));
@@ -1187,7 +1205,7 @@ static int ip_vs_sysctl_defense_mode(ctl_table *ctl, int write,
  */
 struct ip_vs_sysctl_table {
 	struct ctl_table_header *sysctl_header;
-	ctl_table vs_vars[19];
+	ctl_table vs_vars[20];
 	ctl_table vs_dir[2];
 	ctl_table ipv4_dir[2];
 	ctl_table root_dir[2];
@@ -1204,6 +1222,9 @@ static struct ip_vs_sysctl_table ipv4_vs_table = {
           &sysctl_ip_vs_debug_level, sizeof(int), 0644, NULL,
           &proc_dointvec},
 #endif
+         {NET_IPV4_VS_USE_DST, "use_dst",
+          &sysctl_use_dst, sizeof(int), 0644, NULL,
+          &proc_dointvec},
          {NET_IPV4_VS_AMDROPRATE, "am_droprate",
           &sysctl_ip_vs_am_droprate, sizeof(int), 0644, NULL,
           &proc_dointvec},
