@@ -1,7 +1,7 @@
 /*
  * IPVS:        Weighted Round-Robin Scheduling module
  *
- * Version:     $Id: ip_vs_wrr.c,v 1.5 2000/09/06 08:36:36 wensong Exp $
+ * Version:     $Id: ip_vs_wrr.c,v 1.6 2000/11/26 04:35:39 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -99,7 +99,8 @@ static int ip_vs_wrr_max_weight(struct ip_vs_service *svc)
 /*
  *    Weighted Round-Robin Scheduling
  */
-static struct ip_vs_dest* ip_vs_wrr_schedule(struct ip_vs_service *svc)
+static struct ip_vs_dest *
+ip_vs_wrr_schedule(struct ip_vs_service *svc, struct iphdr *iph)
 {
 	struct ip_vs_dest *dest;
 	struct ip_vs_wrr_mark *mark = svc->sched_data;
